"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.normalizeNewFieldContent = exports.applyFieldTypesFromContext = exports.applyTypesFromContext = exports.cursorsForTypedFieldData = exports.cursorsFromContextualData = exports.cursorForTypedData = exports.cursorForTypedTreeData = exports.cursorFromContextualData = exports.isContextuallyTypedNodeDataObject = exports.isArrayLike = exports.arrayLikeMarkerSymbol = exports.getPossibleTypes = exports.getAllowedTypes = exports.getFieldKind = exports.valueSymbol = exports.typeNameSymbol = exports.areCursors = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../core/index.js");
const index_js_2 = require("../util/index.js");
// TODO:
// This module currently is assuming use of default-field-kinds.
// The field kinds should instead come from a view schema registry thats provided somewhere.
const index_js_3 = require("./default-schema/index.js");
const mapTreeCursor_js_1 = require("./mapTreeCursor.js");
const index_js_4 = require("./typed-schema/index.js");
const valueUtilities_js_1 = require("./valueUtilities.js");
/**
 * This library defines a tree data format that can infer its types from context.
 * It can only be used when the schema is known.
 * The format is optimized for ergonomics when the developer knows the schema,
 * and needs to declare or navigate trees.
 *
 * The format defined here is very tolerant to optimize for flexibility of expressing trees:
 * APIs exposing data in this format should likely further constrain what is allowed.
 * For example guarantee which fields and nodes should be inlined, and that types will be required everywhere.
 *
 * This is from Flex tree one which has been deleted and should no longer be used!
 */
/**
 * Check if NewFieldContent is made of {@link ITreeCursor}s.
 *
 * Useful when APIs want to take in tree data in multiple formats, including cursors.
 */
function areCursors(data) {
    if ((0, index_js_1.isCursor)(data)) {
        return true;
    }
    if (Array.isArray(data) && data.length >= 0 && (0, index_js_1.isCursor)(data[0])) {
        return true;
    }
    return false;
}
exports.areCursors = areCursors;
/**
 * String which identifies this code.
 * Targeted at developers: can be used for symbol strings, or other developer targeted strings,
 * like error messages.
 */
const scope = "contextuallyTyped";
/**
 * A symbol for the name of the type of a tree in contexts where string keys are already in use for fields.
 * See {@link TreeNodeSchemaIdentifier}.
 * @internal
 */
exports.typeNameSymbol = Symbol(`${scope}:typeName`);
/**
 * A symbol for the value of a tree node in contexts where string keys are already in use for fields.
 * @internal
 */
exports.valueSymbol = Symbol(`${scope}:value`);
function getFieldKind(fieldSchema) {
    // TODO:
    // This module currently is assuming use of defaultFieldKinds.
    // The field kinds should instead come from a view schema registry thats provided somewhere.
    return index_js_3.fieldKinds.get(fieldSchema.kind.identifier) ?? (0, index_js_2.fail)("missing field kind");
}
exports.getFieldKind = getFieldKind;
/**
 * @returns all allowed child types for `typeSet`.
 */
function getAllowedTypes(schemaData, typeSet) {
    // TODO: Performance: avoid the `Any` case being frequent, possibly with caching in the caller of `getPossibleChildTypes`.
    return typeSet === index_js_4.Any ? new Set(schemaData.nodeSchema.values()) : typeSet;
}
exports.getAllowedTypes = getAllowedTypes;
/**
 * @returns all types, for which the data is schema-compatible.
 */
function getPossibleTypes(context, typeSet, data) {
    // All types allowed by schema
    const allowedTypes = getAllowedTypes(context, typeSet);
    const possibleTypes = [];
    for (const allowed of allowedTypes) {
        if (shallowCompatibilityTest(allowed, data)) {
            possibleTypes.push(allowed);
        }
    }
    return possibleTypes;
}
exports.getPossibleTypes = getPossibleTypes;
/**
 * A symbol used to define a {@link MarkedArrayLike} interface.
 * @internal
 */
exports.arrayLikeMarkerSymbol = Symbol("editable-tree:arrayLikeMarker");
/**
 * Checks the type of a `ContextuallyTypedNodeData`.
 */
function isArrayLike(data) {
    if (typeof data !== "object" || data === null) {
        return false;
    }
    return (data[exports.arrayLikeMarkerSymbol] ===
        true || Array.isArray(data));
}
exports.isArrayLike = isArrayLike;
/**
 * Checks the type of a `ContextuallyTypedNodeData`.
 * @internal
 */
function isContextuallyTypedNodeDataObject(data) {
    return !((0, valueUtilities_js_1.isTreeValue)(data) || isArrayLike(data));
}
exports.isContextuallyTypedNodeDataObject = isContextuallyTypedNodeDataObject;
/**
 * Checks if data might be schema-compatible.
 *
 * @returns false if `data` is incompatible with `type` based on a cheap/shallow check.
 *
 * Note that this may return true for cases where data is incompatible, but it must not return false in cases where the data is compatible.
 */
function shallowCompatibilityTest(schema, data) {
    (0, internal_1.assert)(!areCursors(data), 0x6b1 /* cursors cannot be used as contextually typed data. */);
    (0, internal_1.assert)(data !== undefined, 0x6b2 /* undefined cannot be used as contextually typed data. Use ContextuallyTypedFieldData. */);
    if (schema instanceof index_js_4.LeafNodeSchema) {
        // Reject objects with no value from being leaf nodes.
        // Note that if allowing IFluidHandles without wrapping them in a leaf node object,
        // this (or the above isPrimitiveValue) would have to change.
        if ((0, valueUtilities_js_1.isTreeValue)(data)) {
            return (0, valueUtilities_js_1.allowsValue)(schema.leafValue, data);
        }
        if (data[exports.valueSymbol] === undefined) {
            return false;
        }
    }
    if ((0, valueUtilities_js_1.isTreeValue)(data)) {
        return false;
    }
    if (isArrayLike(data)) {
        if (schema instanceof index_js_4.FlexFieldNodeSchema) {
            const field = schema.getFieldSchema();
            return field.kind.multiplicity === index_js_1.Multiplicity.Sequence;
        }
        else {
            return false;
        }
    }
    if (data[exports.typeNameSymbol] !== undefined) {
        return data[exports.typeNameSymbol] === schema.name;
    }
    // For now, consider all not explicitly typed objects shallow compatible.
    // This will require explicit differentiation in polymorphic cases rather than automatic structural differentiation.
    return true;
}
/**
 * Construct a tree from ContextuallyTypedNodeData.
 *
 * TODO: this should probably be refactored into a `try` function which either returns a Cursor or a SchemaError with a path to the error.
 * @returns a cursor in Nodes mode for a single node containing the provided data.
 * @internal
 */
function cursorFromContextualData(context, typeSet, data) {
    const mapTree = applyTypesFromContext(context, typeSet, data);
    return (0, mapTreeCursor_js_1.cursorForMapTreeNode)(mapTree);
}
exports.cursorFromContextualData = cursorFromContextualData;
/**
 * Strongly typed {@link cursorFromContextualData} for a TreeNodeSchema.
 * @returns a cursor in Nodes mode for a single node containing the provided data.
 * @internal
 */
function cursorForTypedTreeData(context, schema, data) {
    return cursorFromContextualData(context, new Set([schema]), data);
}
exports.cursorForTypedTreeData = cursorForTypedTreeData;
/**
 * Strongly typed {@link cursorFromContextualData} for AllowedTypes.
 * @returns a cursor in Nodes mode for a single node containing the provided data.
 * @internal
 */
function cursorForTypedData(context, schema, data) {
    return cursorFromContextualData(context, (0, index_js_4.allowedTypesSchemaSet)(schema), data);
}
exports.cursorForTypedData = cursorForTypedData;
/**
 * Construct a tree from ContextuallyTypedNodeData.
 * Returns a cursor in Field mode.
 *
 * TODO: this should probably be refactored into a `try` function which either returns a Cursor or a SchemaError with a path to the error.
 */
function cursorsFromContextualData(context, field, data) {
    const mapTrees = applyFieldTypesFromContext(context, field, data);
    return (0, mapTreeCursor_js_1.cursorForMapTreeField)(mapTrees);
}
exports.cursorsFromContextualData = cursorsFromContextualData;
/**
 * Strongly typed {@link cursorsFromContextualData} for a TreeFieldSchema
 * @internal
 */
function cursorsForTypedFieldData(context, schema, data) {
    return cursorsFromContextualData(context, schema, data);
}
exports.cursorsForTypedFieldData = cursorsForTypedFieldData;
/**
 * Construct a MapTree from ContextuallyTypedNodeData.
 *
 * TODO: this should probably be refactored into a `try` function which either returns a MapTree or a SchemaError with a path to the error.
 * TODO: test suite.
 *
 * @remarks
 * This version is only exported as a more testable entry point than `cursorFromContextualData` which keeps the use of `MapTree` as an implementation detail.
 * This should not be reexported from the parent module.
 */
function applyTypesFromContext(context, typeSet, data) {
    const possibleTypes = getPossibleTypes(context.schema, typeSet, data);
    (0, internal_1.assert)(possibleTypes.length !== 0, 0x4d4 /* data incompatible with all types allowed by the schema */);
    (0, internal_1.assert)(possibleTypes.length === 1, 0x4d5 /* data compatible with more than one type allowed by the schema */);
    const schema = possibleTypes[0];
    if (schema instanceof index_js_4.LeafNodeSchema) {
        const value = (0, valueUtilities_js_1.isTreeValue)(data)
            ? data
            : data[exports.valueSymbol];
        (0, internal_1.assert)((0, valueUtilities_js_1.allowsValue)(schema.leafValue, value), 0x4d3 /* unsupported schema for provided primitive */);
        return { value, type: schema.name, fields: new Map() };
    }
    (0, internal_1.assert)(!(0, valueUtilities_js_1.isTreeValue)(data), 0x880 /* leaf value for non leaf */);
    if (schema instanceof index_js_4.FlexFieldNodeSchema) {
        if (isArrayLike(data)) {
            const children = applyFieldTypesFromContext(context, schema.getFieldSchema(), data);
            return {
                value: undefined,
                type: schema.name,
                fields: new Map(children.length > 0 ? [[index_js_1.EmptyKey, children]] : []),
            };
        }
    }
    (0, internal_1.assert)(!isArrayLike(data), 0x881 /* array for non field node */);
    if (schema instanceof index_js_4.FlexMapNodeSchema ||
        schema instanceof index_js_4.FlexObjectNodeSchema ||
        schema instanceof index_js_4.FlexFieldNodeSchema) {
        const fields = new Map();
        for (const key of fieldKeysFromData(data)) {
            (0, internal_1.assert)(!fields.has(key), 0x6b3 /* Keys should not be duplicated */);
            const childSchema = schema.getFieldSchema(key);
            const children = applyFieldTypesFromContext(context, childSchema, data[key]);
            if (children.length > 0) {
                fields.set(key, children);
            }
        }
        if (schema instanceof index_js_4.FlexObjectNodeSchema) {
            for (const key of schema.objectNodeFields.keys()) {
                if (data[key] === undefined) {
                    setFieldForKey(key, context, schema, fields);
                }
            }
        }
        const value = data[exports.valueSymbol];
        (0, internal_1.assert)(value === undefined, 0x4d7 /* provided value not permitted by the schema */);
        return { value, type: schema.name, fields };
    }
    else {
        (0, index_js_2.fail)("unexpected node kind");
    }
}
exports.applyTypesFromContext = applyTypesFromContext;
function setFieldForKey(key, context, schema, fields) {
    const requiredFieldSchema = schema.getFieldSchema(key);
    const multiplicity = getFieldKind(requiredFieldSchema).multiplicity;
    if (multiplicity === index_js_1.Multiplicity.Single && context.fieldSource !== undefined) {
        const fieldGenerator = context.fieldSource(key, requiredFieldSchema.stored);
        if (fieldGenerator !== undefined) {
            const children = fieldGenerator();
            fields.set(key, children);
        }
    }
}
function fieldKeysFromData(data) {
    const keys = Reflect.ownKeys(data).filter((key) => typeof key === "string");
    return keys;
}
/**
 * Construct a MapTree from ContextuallyTypedNodeData.
 *
 * TODO: this should probably be refactored into a `try` function which either returns a MapTree or a SchemaError with a path to the error.
 * TODO: test suite.
 *
 * @remarks
 * This version is only exported as a more testable entry point than `cursorsFromContextualData` which keeps the use of `MapTree` as an implementation detail.
 * This should not be reexported from the parent module.
 */
function applyFieldTypesFromContext(context, field, data) {
    const multiplicity = getFieldKind(field).multiplicity;
    if (data === undefined) {
        (0, internal_1.assert)(multiplicity === index_js_1.Multiplicity.Forbidden || multiplicity === index_js_1.Multiplicity.Optional, 0x4d8 /* `undefined` provided for a field that does not support `undefined` */);
        return [];
    }
    if (multiplicity === index_js_1.Multiplicity.Sequence) {
        (0, internal_1.assert)(isArrayLike(data), 0x4d9 /* expected array for a sequence field */);
        const children = Array.from(data, (child) => applyTypesFromContext(context, field.allowedTypeSet, child));
        return children;
    }
    (0, internal_1.assert)(multiplicity === index_js_1.Multiplicity.Single || multiplicity === index_js_1.Multiplicity.Optional, 0x4da /* single value provided for an unsupported field */);
    return [applyTypesFromContext(context, field.allowedTypeSet, data)];
}
exports.applyFieldTypesFromContext = applyFieldTypesFromContext;
/**
 * Convert NewFieldContent into ITreeCursorSynchronous.
 * The returned cursor will be in Field mode.
 */
function normalizeNewFieldContent(context, schema, content) {
    if (areCursors(content)) {
        if ((0, index_js_2.isReadonlyArray)(content)) {
            (0, internal_1.assert)(getFieldKind(schema).multiplicity === index_js_1.Multiplicity.Sequence || content.length === 1, 0x6b8 /* non-sequence fields can not be provided content that is multiple cursors */);
            // TODO: is there a better way to get a field cursor from an array of node cursors?
            const mapTrees = content.map((c) => (0, mapTreeCursor_js_1.mapTreeFromCursor)(c));
            return (0, mapTreeCursor_js_1.cursorForMapTreeField)(mapTrees);
        }
        if (content.mode === 1 /* CursorLocationType.Fields */) {
            return content;
        }
        // TODO: is there a better way to get a field cursor from a node cursor?
        return (0, mapTreeCursor_js_1.cursorForMapTreeField)([(0, mapTreeCursor_js_1.mapTreeFromCursor)(content)]);
    }
    return cursorsFromContextualData(context, schema, content);
}
exports.normalizeNewFieldContent = normalizeNewFieldContent;
//# sourceMappingURL=contextuallyTyped.js.map