"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.RawObjectNode = exports.objectSchema = exports.setField = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../feature-libraries/index.js");
const proxies_js_1 = require("./proxies.js");
const proxyBinding_js_1 = require("./proxyBinding.js");
const schemaCaching_js_1 = require("./schemaCaching.js");
const schemaTypes_js_1 = require("./schemaTypes.js");
const toMapTree_js_1 = require("./toMapTree.js");
const index_js_2 = require("../util/index.js");
const toFlexSchema_js_1 = require("./toFlexSchema.js");
const rawNode_js_1 = require("./rawNode.js");
/**
 * Caches a {@link SimpleKeyToFlexKeyMap} for a given {@link TreeNodeSchema}.
 */
const simpleKeyToFlexKeyCache = new WeakMap();
/**
 * Caches the mappings from view keys to stored keys for the provided object field schemas in {@link simpleKeyToFlexKeyCache}.
 */
function getOrCreateFlexKeyMapping(nodeSchema, fields) {
    let keyMap = simpleKeyToFlexKeyCache.get(nodeSchema);
    if (keyMap === undefined) {
        keyMap = new Map();
        for (const [viewKey, fieldSchema] of Object.entries(fields)) {
            // Only specify mapping if the stored key differs from the view key.
            // No entry in this map will indicate that the two keys are the same.
            const storedKey = (0, schemaTypes_js_1.getStoredKey)(viewKey, fieldSchema);
            if (viewKey !== storedKey) {
                keyMap.set(viewKey, (0, index_js_2.brand)(storedKey));
            }
        }
        simpleKeyToFlexKeyCache.set(nodeSchema, keyMap);
    }
    return keyMap;
}
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `{}` is used for the target.
 */
function createObjectProxy(schema, allowAdditionalProperties, targetObject = {}) {
    // Performance optimization: cache view key => stored key mapping.
    const flexKeyMap = getOrCreateFlexKeyMapping(schema, schema.info);
    function getFlexKey(viewKey) {
        return flexKeyMap.get(viewKey) ?? (0, index_js_2.brand)(viewKey);
    }
    // To satisfy 'deepEquals' level scrutiny, the target of the proxy must be an object with the same
    // prototype as an object literal '{}'.  This is because 'deepEquals' uses 'Object.getPrototypeOf'
    // as a way to quickly reject objects with different prototype chains.
    //
    // (Note that the prototype of an object literal appears as '[Object: null prototype] {}', not because
    // the prototype is null, but because the prototype object itself has a null prototype.)
    // TODO: Although the target is an object literal, it's still worthwhile to try experimenting with
    // a dispatch object to see if it improves performance.
    const proxy = new Proxy(targetObject, {
        get(target, viewKey) {
            const flexKey = getFlexKey(viewKey);
            const field = (0, proxyBinding_js_1.getFlexNode)(proxy).tryGetField(flexKey);
            if (field !== undefined) {
                return (0, proxies_js_1.getProxyForField)(field);
            }
            // Pass the proxy as the receiver here, so that any methods on the prototype receive `proxy` as `this`.
            return Reflect.get(target, viewKey, proxy);
        },
        set(target, viewKey, value) {
            const flexNode = (0, proxyBinding_js_1.getFlexNode)(proxy);
            const flexNodeSchema = flexNode.schema;
            (0, internal_1.assert)(flexNodeSchema instanceof index_js_1.FlexObjectNodeSchema, 0x888 /* invalid schema */);
            const hasField = viewKey in schema.info;
            if (!hasField) {
                return allowAdditionalProperties ? Reflect.set(target, viewKey, value) : false;
            }
            (0, internal_1.assert)(typeof viewKey === "string", 0x7e1 /* invalid key */);
            const flexKey = getFlexKey(viewKey);
            const field = flexNode.getBoxed(flexKey);
            const simpleNodeFields = schema.info;
            if (simpleNodeFields[viewKey] === undefined) {
                (0, index_js_2.fail)(`Field key '${viewKey}' not found in schema.`);
            }
            const simpleFieldSchema = (0, schemaCaching_js_1.getSimpleFieldSchema)(field.schema, schema.info[viewKey]);
            setField(field, simpleFieldSchema, value);
            return true;
        },
        has: (target, viewKey) => {
            const fields = schema.info;
            return (fields[viewKey] !== undefined ||
                (allowAdditionalProperties ? Reflect.has(target, viewKey) : false));
        },
        ownKeys: (target) => {
            const fields = schema.info;
            return [
                ...Object.keys(fields),
                ...(allowAdditionalProperties ? Reflect.ownKeys(target) : []),
            ];
        },
        getOwnPropertyDescriptor: (target, viewKey) => {
            const flexKey = getFlexKey(viewKey);
            const field = (0, proxyBinding_js_1.getFlexNode)(proxy).tryGetField(flexKey);
            if (field === undefined) {
                return allowAdditionalProperties
                    ? Reflect.getOwnPropertyDescriptor(target, viewKey)
                    : undefined;
            }
            const p = {
                value: (0, proxies_js_1.getProxyForField)(field),
                writable: true,
                enumerable: true,
                configurable: true, // Must be 'configurable' if property is absent from proxy target.
            };
            return p;
        },
    });
    return proxy;
}
function setField(field, simpleFieldSchema, value) {
    switch (field.schema.kind) {
        case index_js_1.FieldKinds.required:
        case index_js_1.FieldKinds.optional: {
            const typedField = field;
            const content = (0, proxies_js_1.prepareContentForInsert)(value, field.context.forest);
            const cursor = (0, toMapTree_js_1.cursorFromNodeData)(content, simpleFieldSchema.allowedTypes);
            typedField.content = cursor;
            break;
        }
        default:
            (0, index_js_2.fail)("invalid FieldKind");
    }
}
exports.setField = setField;
/**
 * Define a {@link TreeNodeSchema} for a {@link TreeObjectNode}.
 *
 * @param name - Unique identifier for this schema within this factory's scope.
 * @param fields - Schema for fields of the object node's schema. Defines what children can be placed under each key.
 */
function objectSchema(base) {
    // Ensure no collisions between final set of view keys, and final set of stored keys (including those
    // implicitly derived from view keys)
    assertUniqueKeys(base.identifier, base.info);
    class schema extends base {
        constructor(input) {
            super(input);
            // Differentiate between the following cases:
            //
            // Case 1: Direct construction (POJO emulation)
            //
            //     const Foo = schemaFactory.object("Foo", {bar: schemaFactory.number});
            //
            //     assert.deepEqual(new Foo({ bar: 42 }), { bar: 42 },
            //		   "Prototype chain equivalent to POJO.");
            //
            // Case 2: Subclass construction (Customizable Object)
            //
            // 	   class Foo extends schemaFactory.object("Foo", {bar: schemaFactory.number}) {}
            //
            // 	   assert.notDeepEqual(new Foo({ bar: 42 }), { bar: 42 },
            // 	       "Subclass prototype chain differs from POJO.");
            //
            // In Case 1 (POJO emulation), the prototype chain match '{}' (proxyTarget = undefined)
            // In Case 2 (Customizable Object), the prototype chain include the user's subclass (proxyTarget = this)
            const customizable = this.constructor !== schema;
            const proxyTarget = customizable ? this : undefined;
            const flexSchema = (0, toFlexSchema_js_1.getFlexSchema)(this.constructor);
            (0, internal_1.assert)(flexSchema instanceof index_js_1.FlexObjectNodeSchema, "invalid flex schema");
            const flexNode = (0, index_js_1.isFlexTreeNode)(input)
                ? input
                : new RawObjectNode(flexSchema, copyContent(flexSchema.name, input));
            const proxy = createObjectProxy(this.constructor, customizable, proxyTarget);
            (0, proxyBinding_js_1.setFlexNode)(proxy, flexNode);
            return proxy;
        }
    }
    return schema;
}
exports.objectSchema = objectSchema;
/**
 * The implementation of an object node created by {@link createRawNode}.
 */
class RawObjectNode extends rawNode_js_1.RawTreeNode {
    get localNodeKey() {
        throw (0, rawNode_js_1.rawError)("Reading local node keys");
    }
}
exports.RawObjectNode = RawObjectNode;
/**
 * Ensures that the set of view keys in the schema is unique.
 * Also ensure that the final set of stored keys (including those implicitly derived from view keys) is unique.
 * @throws Throws a `UsageError` if either of the key uniqueness invariants is violated.
 */
function assertUniqueKeys(schemaName, fields) {
    // Verify that there are no duplicates among the explicitly specified stored keys.
    const explicitStoredKeys = new Set();
    for (const schema of Object.values(fields)) {
        const storedKey = (0, schemaTypes_js_1.getExplicitStoredKey)(schema);
        if (storedKey === undefined) {
            continue;
        }
        if (explicitStoredKeys.has(storedKey)) {
            throw new internal_2.UsageError(`Duplicate stored key "${storedKey}" in schema "${schemaName}". Stored keys must be unique within an object schema.`);
        }
        explicitStoredKeys.add(storedKey);
    }
    // Verify that there are no duplicates among the derived
    // (including those implicitly derived from view keys) stored keys.
    const derivedStoredKeys = new Set();
    for (const [viewKey, schema] of Object.entries(fields)) {
        const storedKey = (0, schemaTypes_js_1.getStoredKey)(viewKey, schema);
        if (derivedStoredKeys.has(storedKey)) {
            throw new internal_2.UsageError(`Stored key "${storedKey}" in schema "${schemaName}" conflicts with a property key of the same name, which is not overridden by a stored key. The final set of stored keys in an object schema must be unique.`);
        }
        derivedStoredKeys.add(storedKey);
    }
}
function copyContent(typeName, content) {
    const copy = { ...content };
    (0, proxies_js_1.markContentType)(typeName, copy);
    return copy;
}
//# sourceMappingURL=objectNode.js.map